/*
    CREATE MATHJS HIGHLIGHT LANGUAGE
*/



function mathjs(hljs) {

    function either(args) {
        return "(" + args.join("|") + ")";
    }

    function keywords(args) {
        return "\\b" + either(args) + "\\b";
    }

    const OPERATORS = {
        scope: "punctuation",
        begin: /[=+\-*/.^?!%'<>:~]+/,
    };

    const OPERATOR_KEYWORDS = {
        scope: "punctuation",
        begin: keywords(["to", "in", "and", "not", "or", "xor", "mod"]),
    };

    const NUMBER_KEYWORDS = {
        scope: "number",
        begin: keywords(["true", "false", "end"]),
    };

    const UNIT_LIST = Object.keys(math.Unit.UNITS).sort((a, b) => b.length - a.length);

    const PREFIXED_UNIT_LIST = UNIT_LIST.map(e => {
            const prefixes = Object.keys(math.Unit.UNITS[e].prefixes).filter(e => e);
            const prefix_re = prefixes.length ? either(prefixes) + "?" : "";
            return prefix_re + e;
    });

    const UNITS = {
        scope: "class",
        begin: keywords(PREFIXED_UNIT_LIST),
    };

    const MATRIX = {
        scope: "bullet",
        begin: /[\[\],;]/,
    };

    const FUNCTION_LIST = Object.getOwnPropertyNames(math).filter(e => typeof math[e] === "function");

    const FUNCTIONS = {
        scope: "title",
        begin: keywords(FUNCTION_LIST),
    };

    const VARIABLE = {
        scope: "variable",
        begin: /[a-zA-Z_$][a-zA-Z0-9_$]*/,
    };

    return {
        name: "mathjs",
        contains: [
            hljs.HASH_COMMENT_MODE,
            hljs.C_NUMBER_MODE,
            NUMBER_KEYWORDS,
            OPERATORS,
            OPERATOR_KEYWORDS,
            UNITS,
            MATRIX,
            FUNCTIONS,
            VARIABLE,
        ],
    };
}

hljs.registerLanguage("mathjs", mathjs);










/*
    DEFINE CODEJAR GLOBAL VARIABLES
*/


function withLineNumbers(highlight, options = {}) {
    const opts = Object.assign({ class: "codejar-linenumbers", wrapClass: "codejar-wrap" }, options);
    let lineNumbers;
    return function (editor) {
        highlight(editor);
        if (!lineNumbers) {
            lineNumbers = withLineNumbersInit(editor, opts);
            editor.addEventListener("scroll", () => lineNumbers.style.top = `-${editor.scrollTop}px`);
        }
        const code = editor.textContent || "";
        const linesCount = code.replace(/\n+$/, "\n").split("\n").length + 1;
        let text = "";
        for (let i = 1; i < linesCount; i++) {
            text += `${i}\n`;
        }
        lineNumbers.innerText = text;
    };
}
function withLineNumbersInit(editor, opts) {
    const css = getComputedStyle(editor);
    const wrap = document.createElement("div");
    wrap.className = opts.wrapClass;
    wrap.style.position = "relative";
    const gutter = document.createElement("div");
    gutter.className = opts.class;
    wrap.appendChild(gutter);
    // Add own styles
    gutter.style.position = "absolute";
    gutter.style.top = "0px";
    gutter.style.left = "0px";
    gutter.style.bottom = "0px";
    gutter.style.overflow = "hidden";
    // Copy editor styles
    gutter.style.fontFamily = css.fontFamily;
    gutter.style.fontSize = css.fontSize;
    gutter.style.lineHeight = css.lineHeight;
    gutter.style.paddingTop = css.paddingTop;
    gutter.style.paddingLeft = css.paddingLeft;
    gutter.style.borderTopLeftRadius = css.borderTopLeftRadius;
    gutter.style.borderBottomLeftRadius = css.borderBottomLeftRadius;
    // Add line numbers
    const lineNumbers = document.createElement("div");
    lineNumbers.style.position = "relative";
    lineNumbers.style.top = "0px";
    gutter.appendChild(lineNumbers);
    // Tweak editor styles
    editor.style.whiteSpace = "pre";
    // Swap editor with a wrap
    editor.parentNode.insertBefore(wrap, editor);
    wrap.appendChild(editor);
    return lineNumbers;
}



window.editor = CodeJar(document.querySelector("#input"), withLineNumbers(hljs.highlightElement));
window.results = CodeJar(document.querySelector("#output"), withLineNumbers(hljs.highlightElement));

document.querySelector("#output").setAttribute("contenteditable", "false");










/*
    ANDROID FUNCTIONS
*/



function get_note() {
    if(window.editor) {
        let code = window.editor.toString().trim();
        if(code == "# Type here...") return "";
        return code;
    }
    return "";
}



function share_note() {

    if(!window.editor) return "";

    let input = window.editor.toString();

    let output = [];
    let scope = make_scope();
    let index = 0;

    for(const line of input.split("\n")) {
        index += 1;
        let output_line = "";
        if(line.trim()) {
            if(line.startsWith("#")) {
                output_line = "";
            } else {
                try {
                    //const r = math.evaluate(line, scope);
                    const r = math.evaluate(`ans = ${line}`, scope);
                    if(r || r === 0) {
                        if(!r.doc) output_line = math.format(r, 14);
                        output_line = output_line.replace(/\n/g, "\\n");
                    }
                } catch(e) {
                    output_line = e.toString().replace(/\n/g, "\\n");
                }
            }
        }
        if(output_line == "") output.push(line);
        else output.push(line + "  # " + output_line);
    }

    return output.join("\n");

}










/*
    MATHPAD
*/



// Delay before execution (to not eval on each key tap)

const delay = 250;



// Code on update

var timer;
editor.onUpdate(code => {
    let cursor = editor.save();
    if(cursor.start > 1 && cursor.start == cursor.end && cursor.dir == "->") {
        let pos = cursor.start;
        let last_char_1 = code.charAt(pos-1);
        let last_char_2 = code.charAt(pos-2);
        if(last_char_2 == "\n" && ["-","+","*","/","^"].includes(last_char_1)) {
            code = code.substring(0, pos-1) + "ans" + last_char_1 + code.substring(pos);
            editor.updateCode(code);
            cursor.start += 3;
            cursor.end   += 3;
            editor.restore(cursor);
        }
    }
    clearTimeout(timer);
    timer = setTimeout(eval_math, delay, code);
});



// Drag and drop files

document.querySelector("#input").addEventListener("drop", (event) => {
    event.preventDefault();
    const file = event.dataTransfer.items[0].getAsFile();
    file.text().then(e => editor.updateCode(e));
});



// Make math scope

function make_scope() {
    
    let scope = {};

    math.import({
        and: function(arg1) {
            var o1 = !(!arg1);
            for(var i = 1; i < arguments.length; i++) {
                var o2 = !(!arguments[i]);
                o1 = o1 && o2;
            }
            return !(!o1);
        },
        xor: function(arg1) {
            var o1 = !(!arg1);
            for(var i = 1; i < arguments.length; i++) {
                var o2 = !(!arguments[i]);
                o1 = o1 ^ o2;
            }
            return !(!o1);
        },
        or: function(arg1) {
            var o1 = !(!arg1);
            for(var i = 1; i < arguments.length; i++) {
                var o2 = !(!arguments[i]);
                o1 = o1 || o2;
            }
            return !(!o1);
        }
    }, {
        override: true
    });

    math.evaluate(`sinr(x) = sin(x)`, scope);
    math.evaluate(`cosr(x) = cos(x)`, scope);
    math.evaluate(`tanr(x) = tan(x)`, scope);
    math.evaluate(`cotr(x) = cot(x)`, scope);
    math.evaluate(`secr(x) = sec(x)`, scope);
    math.evaluate(`cscr(x) = csc(x)`, scope);
    math.evaluate(`asinr(x) = asin(x)`, scope);
    math.evaluate(`acosr(x) = acos(x)`, scope);
    math.evaluate(`atanr(x) = atan(x)`, scope);
    
    math.evaluate(`sind(x) = sin(x * pi / 180)`, scope);
    math.evaluate(`cosd(x) = cos(x * pi / 180)`, scope);
    math.evaluate(`tand(x) = tan(x * pi / 180)`, scope);
    math.evaluate(`cotd(x) = cot(x * pi / 180)`, scope);
    math.evaluate(`secd(x) = sec(x * pi / 180)`, scope);
    math.evaluate(`cscd(x) = csc(x * pi / 180)`, scope);
    math.evaluate(`asind(x) = asin(x) / pi * 180`, scope);
    math.evaluate(`acosd(x) = acos(x) / pi * 180`, scope);
    math.evaluate(`atand(x) = atan(x) / pi * 180`, scope);
    
    math.evaluate(`sing(x) = sin(x * pi / 200)`, scope);
    math.evaluate(`cosg(x) = cos(x * pi / 200)`, scope);
    math.evaluate(`tang(x) = tan(x * pi / 200)`, scope);
    math.evaluate(`cotg(x) = cot(x * pi / 200)`, scope);
    math.evaluate(`secg(x) = sec(x * pi / 200)`, scope);
    math.evaluate(`cscg(x) = csc(x * pi / 200)`, scope);
    math.evaluate(`asing(x) = asin(x) / pi * 200`, scope);
    math.evaluate(`acosg(x) = acos(x) / pi * 200`, scope);
    math.evaluate(`atang(x) = atan(x) / pi * 200`, scope);
    
    math.evaluate(`avg = mean`, scope);
    
    math.evaluate(`log(x) = log10(x)`, scope);
    math.evaluate(`ln(x) = log(x) / log(e)`, scope);
    
    math.evaluate(`if(x, y, z) = x ? y : z`, scope);

    return scope;

}



// Evaluate each line

function eval_math(input) {

    let output = [];
    let scope = make_scope();
    let index = 0;

    for(const line of input.split("\n")) {
        index += 1;
        let output_line = "";
        if(line.trim()) {
            if(line.startsWith("#")) {
                output_line = line;
            } else {
                try {
                    //const r = math.evaluate(line, scope);
                    const r = math.evaluate(`ans = ${line}`, scope);
                    if(r || r === 0) {
                        if(!r.doc) output_line = math.format(r, 14);
                        output_line = output_line.replace(/\n/g, "\\n");
                    }
                } catch(e) {
                    output_line = e.toString().replace(/\n/g, "\\n");
                }
            }
        }
        output.push(output_line);
    }

    results.updateCode(output.join("\n"));

}

hljs.configure({ ignoreUnescapedHTML: true });



// Start

let params = new URLSearchParams(window.location.search);
let code = params.get("note");
code = atob(code);
if(code == "") code = "# Type here...";
//if(code != "") code = atob(code);
//else code = `# Expressions\n\n1.2 / (3.3 + 1.7)\n\n# Variables\n\na = 5.08 + 2\nb = a + 4.87\n\n# Functions\n\nf(x, y) = x ^ y\nf(2, 3)\n\n# Type here...`;
editor.updateCode(code);
eval_math(editor.toString());
